-- * Imports

-- ** Stdlib

open import Function

open import Relation.Binary.PropositionalEquality

open import Data.Product

-- ** Local

open import Pow
open import IProp

open import Container.Container
open import Container.Fixpoint
open import Container.Initiality

open import Ornament.Ornament
open import Ornament.OrnamentalAlgebra

module Ornament.AlgebraicOrnament {O : Set}{X : O → Set}
                                  (φ : O ▸ O) 
                                  (α : ⟦ φ ⟧ X ⟶ X) where

-- * The Algebraic Ornament

-- Given an algebra on a container, we can compute an ornament which
-- indices respects the algebra.


-- The (trivial) forget:
forget : Σ O X → O
forget = proj₁

-- Ornament φ with the algebra α:
algebraicOrn : Ornament1 φ (Σ O X) forget
algebraicOrn = record { extend = extendAlgOrn
                      ; refine = refineAlgOrn
                      ; coherence⊢ = coherenceAlgOrn }

         where extendAlgOrn : (o+ : Σ O X) → S φ (forget o+) → Set
               extendAlgOrn o+ sh = Σ[ c ∶ ⟦ φ ⟧ X (forget o+) ] 
                                   (proj₁ c ≡ sh) × 
                                   (α c ≡ (proj₂ o+))
       
               refineAlgOrn : {o+ : Σ O X}{sh : S φ (forget o+)}
                              (e : extendAlgOrn o+ sh) → P φ sh → Σ O X
               refineAlgOrn ((sh , next) , refl , q) p = n φ p , next p

               coherenceAlgOrn : {o+ : Σ O X}(sh : S φ (forget o+))
                                 (e : extendAlgOrn o+ sh)
                                 (p : P φ sh) →
                                   proj₁ (refineAlgOrn e p) ≡ n φ p
               coherenceAlgOrn .sh ((sh , next) , refl , q) p = refl

algebraicOrnament : Σ O X ▸ Σ O X
algebraicOrnament = ⟨ algebraicOrn ⟩o

-- * Bob's Theorem 5: μF^α ≡ Σ_α μF

φ+ : Σ O X ▸ Σ O X
φ+ = algebraicOrnament

-- ** First direction: Σ_α μF → μF^α

makeAlgOrn : {o : O} → (x : μ φ o) → μ φ+ (o , foldCont α x)
makeAlgOrn x = < inductionCont φ 
                          (\{o} x → ⟦ φ+ ⟧ (μ φ+) (o , foldCont α x))
                          help 
                          x >
          where help : {o : O}
                       (xs : ⟦ φ ⟧ (μ φ) o) →
                       □ φ (μ φ) (\{o} x → ⟦ φ+ ⟧ (μ φ+) (o , foldCont α x)) xs →
                       ⟦ φ+ ⟧ (μ φ+) (o , foldCont {P = X} α < xs >)
                help (sh , pos) m = ((sh , ((sh , \p → foldCont α (pos p))) , refl , refl) , <_> ∘ m )


-- ** Second direction: μF^α → Σ_α μF

-- We proceed in two steps:
--   - first, we 'forget' μF^α to μF
--   - second, we show that the resulting μF satisfies α, 
--     (aka. the OAAO theorem)

-- *** μF^α -> μF
 
forgetAlgOrn : {o : O}{x : X o} → μ algebraicOrnament (o , x) → μ φ o
forgetAlgOrn = forgetOrn algebraicOrn

-- *** OAAO theorem:

-- First, we need a well known result on fusing a natural
-- transformation into a fold:

-- TODO This lemma should probably rely on FuseNT

lemma : {i : O}{j : X i}
          (x : μ algebraicOrnament (i , j)) →
        ⊢ foldCont α (forgetAlgOrn x) ≡ foldCont (α ∘ forgetOrnament algebraicOrn X) x
lemma < (.sh , (sh , p) , refl , q) , pos > = 
      pure (cong (λ p → α (sh , p))) ⊛ 
                 extensionality (λ p → lemma (pos p))

-- The OAAO theorem (denoted OAAO⊢) states that what we have indexed the
-- container with is really what we would computed by running the
-- algebra over the raw structure: we have turned an invariant into an
-- index.

OAAO⊢ :  {i : O}{j : X i}
          (x : μ algebraicOrnament (i , j)) → ⊢ foldCont α (forgetAlgOrn x) ≡ j
OAAO⊢ x = pure trans ⊛ 
               lemma x ⊛ 
               initialityCata⊢ (α ∘ forgetOrnament algebraicOrn X) x extractIndex help
  where 
        extractIndex : {ij : Σ O X} → μ algebraicOrnament ij → X (proj₁ ij)
        extractIndex {ij} x = proj₂ ij

        help : {o+ : Σ O X}
                  (xs : ⟦ algebraicOrnament ⟧ (μ algebraicOrnament) o+) → 
                  ⊢ extractIndex < xs > ≡ (α ∘ forgetOrnament algebraicOrn X
                                           ∘ ⟦ algebraicOrnament ⟧Morph (λ {ox} → extractIndex {ox})) xs
        help {o , x} ((.sh , (sh , next) , refl , q) , pos) with x | q
        ... | .(α (sh , next)) | refl = pure (cong (\x → α (sh , x)) refl)



splitAlgOrn : {o : O}{x : X o} → μ φ+ (o , x) → Σ[ t ∶ μ φ o ] ⊢ foldCont α t ≡ x
splitAlgOrn x = forgetAlgOrn x , OAAO⊢ x

-- We have that spliAlgOrn = makeAlgOrn⁻¹


-- * Outro

-- Local Variables:
-- mode: outline-minor
-- outline-regexp: "-- [*\f]+"
-- outline-level: outline-level
-- End:                      
